"""Tests for SQL-backed memory backend with per-user scoping."""

from __future__ import annotations

import pytest

from agent_gateway.memory.backends.sql import SqlMemoryBackend, SqlMemoryRepository
from agent_gateway.memory.domain import MemoryRecord, MemoryType
from agent_gateway.persistence.backends.sqlite import SqliteBackend


@pytest.fixture
async def backend(tmp_path) -> SqliteBackend:
    db_path = tmp_path / "test.db"
    b = SqliteBackend(path=str(db_path))
    await b.initialize()
    yield b
    await b.dispose()


@pytest.fixture
def sql_memory(backend: SqliteBackend) -> SqlMemoryBackend:
    return SqlMemoryBackend(backend._session_factory)


@pytest.fixture
def repo(sql_memory: SqlMemoryBackend) -> SqlMemoryRepository:
    return sql_memory.memory_repo


def _make_record(
    memory_id: str = "m1",
    agent_id: str = "agent-1",
    content: str = "test fact",
    user_id: str | None = None,
    memory_type: MemoryType = MemoryType.SEMANTIC,
    importance: float = 0.5,
) -> MemoryRecord:
    return MemoryRecord(
        id=memory_id,
        agent_id=agent_id,
        content=content,
        user_id=user_id,
        memory_type=memory_type,
        importance=importance,
    )


class TestSqlMemoryCRUD:
    async def test_save_and_get(self, repo: SqlMemoryRepository) -> None:
        record = _make_record()
        await repo.save(record)
        fetched = await repo.get("agent-1", "m1")
        assert fetched is not None
        assert fetched.content == "test fact"

    async def test_get_wrong_agent(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record())
        assert await repo.get("wrong-agent", "m1") is None

    async def test_upsert_updates_existing(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(content="original"))
        await repo.save(_make_record(content="updated"))
        fetched = await repo.get("agent-1", "m1")
        assert fetched is not None
        assert fetched.content == "updated"

    async def test_delete(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record())
        assert await repo.delete("agent-1", "m1") is True
        assert await repo.get("agent-1", "m1") is None

    async def test_delete_wrong_agent(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record())
        assert await repo.delete("wrong-agent", "m1") is False

    async def test_delete_nonexistent(self, repo: SqlMemoryRepository) -> None:
        assert await repo.delete("agent-1", "nope") is False


class TestSqlMemoryUserScoping:
    async def test_global_memory_has_null_user_id(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(user_id=None))
        records = await repo.list_memories("agent-1", user_id=None)
        assert len(records) == 1
        assert records[0].user_id is None

    async def test_user_memory_isolated(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="g1", user_id=None, content="global"))
        await repo.save(_make_record(memory_id="u1", user_id="user-a", content="user-a fact"))
        await repo.save(_make_record(memory_id="u2", user_id="user-b", content="user-b fact"))

        # User A should only see their own + global
        user_a = await repo.list_memories("agent-1", user_id="user-a", include_global=True)
        assert len(user_a) == 2
        contents = {r.content for r in user_a}
        assert "global" in contents
        assert "user-a fact" in contents
        assert "user-b fact" not in contents

    async def test_user_only_excludes_global(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="g1", user_id=None, content="global"))
        await repo.save(_make_record(memory_id="u1", user_id="user-a", content="user-a fact"))

        results = await repo.list_memories("agent-1", user_id="user-a", include_global=False)
        assert len(results) == 1
        assert results[0].content == "user-a fact"

    async def test_global_only_when_no_user(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="g1", user_id=None))
        await repo.save(_make_record(memory_id="u1", user_id="user-a"))

        results = await repo.list_memories("agent-1", user_id=None)
        assert len(results) == 1
        assert results[0].user_id is None


class TestSqlMemorySearch:
    async def test_search_finds_matching(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="m1", content="user prefers dark mode"))
        await repo.save(_make_record(memory_id="m2", content="user likes python"))

        results = await repo.search("agent-1", "dark mode")
        assert len(results) >= 1
        assert results[0].record.content == "user prefers dark mode"

    async def test_search_respects_user_scope(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="g1", user_id=None, content="global dark mode"))
        await repo.save(_make_record(memory_id="u1", user_id="user-a", content="user dark mode"))

        # Search as user-a — should find both
        results = await repo.search("agent-1", "dark mode", user_id="user-a", include_global=True)
        assert len(results) == 2

        # User-only — should find just the user memory
        results = await repo.search("agent-1", "dark mode", user_id="user-a", include_global=False)
        assert len(results) == 1
        assert results[0].record.user_id == "user-a"

    async def test_search_boosts_user_memories(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="g1", user_id=None, content="dark mode setting"))
        await repo.save(
            _make_record(memory_id="u1", user_id="user-a", content="dark mode setting")
        )

        results = await repo.search("agent-1", "dark mode", user_id="user-a", include_global=True)
        assert len(results) == 2
        # User memory should be ranked first (boosted)
        assert results[0].record.user_id == "user-a"

    async def test_search_empty_query_words(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record())
        results = await repo.search("agent-1", "")
        assert results == []

    async def test_search_by_memory_type(self, repo: SqlMemoryRepository) -> None:
        await repo.save(
            _make_record(memory_id="s1", content="fact", memory_type=MemoryType.SEMANTIC)
        )
        await repo.save(
            _make_record(memory_id="e1", content="event fact", memory_type=MemoryType.EPISODIC)
        )

        results = await repo.search("agent-1", "fact", memory_type=MemoryType.SEMANTIC)
        assert len(results) == 1
        assert results[0].record.memory_type == MemoryType.SEMANTIC


class TestSqlMemoryBulkOps:
    async def test_delete_all_no_filter(self, repo: SqlMemoryRepository) -> None:
        """delete_all with user_id=None removes ALL memories for the agent."""
        for i in range(5):
            await repo.save(_make_record(memory_id=f"m{i}", user_id=None))
        await repo.save(_make_record(memory_id="u1", user_id="user-a"))

        deleted = await repo.delete_all("agent-1", user_id=None)
        assert deleted == 6

    async def test_delete_all_user(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="g1", user_id=None))
        for i in range(3):
            await repo.save(_make_record(memory_id=f"u{i}", user_id="user-a"))

        deleted = await repo.delete_all("agent-1", user_id="user-a")
        assert deleted == 3
        assert await repo.get("agent-1", "g1") is not None

    async def test_count_all(self, repo: SqlMemoryRepository) -> None:
        """count with user_id=None returns total count for the agent."""
        for i in range(4):
            await repo.save(_make_record(memory_id=f"m{i}", user_id=None))
        await repo.save(_make_record(memory_id="u1", user_id="user-a"))

        assert await repo.count("agent-1", user_id=None) == 5

    async def test_count_user(self, repo: SqlMemoryRepository) -> None:
        await repo.save(_make_record(memory_id="g1", user_id=None))
        for i in range(3):
            await repo.save(_make_record(memory_id=f"u{i}", user_id="user-a"))

        assert await repo.count("agent-1", user_id="user-a") == 3


class TestSqlMemoryBackend:
    async def test_initialize_dispose(self, sql_memory: SqlMemoryBackend) -> None:
        await sql_memory.initialize()
        await sql_memory.dispose()

    def test_memory_repo_property(self, sql_memory: SqlMemoryBackend) -> None:
        assert isinstance(sql_memory.memory_repo, SqlMemoryRepository)



"""Tests for the composite auth provider."""

from __future__ import annotations

from agent_gateway.auth.domain import AuthResult
from agent_gateway.auth.providers.composite import CompositeProvider


class _AlwaysOk:
    async def authenticate(self, token: str) -> AuthResult:
        return AuthResult.ok(subject="ok-provider", scopes=["*"], method="test")

    async def close(self) -> None:
        pass


class _AlwaysDenied:
    def __init__(self, error: str = "denied") -> None:
        self._error = error

    async def authenticate(self, token: str) -> AuthResult:
        return AuthResult.denied(self._error)

    async def close(self) -> None:
        pass


class _TrackingProvider:
    def __init__(self) -> None:
        self.closed = False

    async def authenticate(self, token: str) -> AuthResult:
        return AuthResult.denied("tracking")

    async def close(self) -> None:
        self.closed = True


class TestCompositeProvider:
    async def test_first_match_wins(self) -> None:
        provider = CompositeProvider([_AlwaysOk(), _AlwaysDenied()])
        result = await provider.authenticate("any-token")
        assert result.authenticated is True
        assert result.subject == "ok-provider"

    async def test_fallback_to_second(self) -> None:
        provider = CompositeProvider([_AlwaysDenied(), _AlwaysOk()])
        result = await provider.authenticate("any-token")
        assert result.authenticated is True

    async def test_all_fail_returns_last_error(self) -> None:
        provider = CompositeProvider(
            [
                _AlwaysDenied("error-1"),
                _AlwaysDenied("error-2"),
            ]
        )
        result = await provider.authenticate("any-token")
        assert result.authenticated is False
        assert result.error == "error-2"

    async def test_empty_providers(self) -> None:
        provider = CompositeProvider([])
        result = await provider.authenticate("any-token")
        assert result.authenticated is False
        assert "No auth providers configured" in result.error

    async def test_close_disposes_all(self) -> None:
        p1 = _TrackingProvider()
        p2 = _TrackingProvider()
        provider = CompositeProvider([p1, p2])  # type: ignore[list-item]
        await provider.close()
        assert p1.closed
        assert p2.closed



"""Schedule management endpoints — list, pause, resume, trigger, create, delete schedules."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from fastapi import APIRouter, Depends, Path, Request
from fastapi.responses import JSONResponse
from pydantic import BaseModel, Field

from agent_gateway.api.errors import error_response, not_found
from agent_gateway.api.models import ScheduleDetailInfo, ScheduleInfo
from agent_gateway.api.openapi import build_responses
from agent_gateway.api.routes.base import GatewayAPIRoute
from agent_gateway.auth.scopes import RequireScope

if TYPE_CHECKING:
    from agent_gateway.gateway import Gateway

logger = logging.getLogger(__name__)

router = APIRouter(route_class=GatewayAPIRoute)

_SCHEDULE_ID_PATTERN = r"^[a-zA-Z0-9_.:/-]+$"


class CreateScheduleRequest(BaseModel):
    """Request body for creating an admin schedule."""

    agent_id: str = Field(..., min_length=1, max_length=256)
    name: str = Field(..., min_length=1, max_length=128, pattern=r"^[a-zA-Z0-9_.\-]+$")
    cron_expr: str = Field(..., min_length=9, max_length=128)
    message: str = Field(..., min_length=1, max_length=102_400)
    instructions: str | None = Field(None, max_length=4000)
    input: dict[str, Any] = Field(default_factory=dict)
    timezone: str = Field("UTC", max_length=64)
    enabled: bool = True


@router.get(
    "/schedules",
    response_model=list[ScheduleInfo],
    summary="List schedules",
    description="List all registered cron schedules.",
    tags=["Schedules"],
    responses=build_responses(auth=True),
    dependencies=[Depends(RequireScope("schedules:read"))],
)
async def list_schedules(request: Request) -> list[ScheduleInfo]:
    """List all registered schedules."""
    gw: Gateway = request.app
    schedules = await gw.list_schedules()
    return [ScheduleInfo(**s) for s in schedules]


@router.post(
    "/schedules",
    summary="Create admin schedule",
    description="Create a new admin-managed schedule for an agent.",
    tags=["Schedules"],
    responses={
        201: {"description": "Schedule created."},
        400: {"description": "Invalid schedule configuration."},
        409: {"description": "Schedule with this name already exists."},
        **build_responses(auth=True, not_found=True),
    },
    dependencies=[Depends(RequireScope("schedules:manage"))],
)
async def create_schedule(
    request: Request,
    body: CreateScheduleRequest,
) -> JSONResponse:
    """Create a new admin schedule."""
    from agent_gateway.exceptions import ScheduleConflictError, ScheduleValidationError

    gw: Gateway = request.app
    if gw.scheduler is None:
        return error_response(404, "scheduler_not_active", "Scheduler is not active")

    # Validate agent exists at route level as UX guard
    if body.agent_id not in gw.agents:
        return not_found("agent", body.agent_id)

    try:
        schedule_id = await gw.create_admin_schedule(
            agent_id=body.agent_id,
            name=body.name,
            cron_expr=body.cron_expr,
            message=body.message,
            instructions=body.instructions,
            input_data=body.input,
            timezone=body.timezone,
            enabled=body.enabled,
        )
    except ScheduleValidationError as e:
        return error_response(400, "schedule_validation_error", str(e))
    except ScheduleConflictError as e:
        return error_response(409, "schedule_conflict", str(e))

    return JSONResponse(
        status_code=201,
        content={"status": "created", "schedule_id": schedule_id},
    )


@router.get(
    "/schedules/{schedule_id:path}",
    response_model=ScheduleDetailInfo,
    summary="Get schedule details",
    description="Get detailed information about a specific schedule.",
    tags=["Schedules"],
    responses=build_responses(auth=True, not_found=True),
    dependencies=[Depends(RequireScope("schedules:read"))],
)
async def get_schedule(
    request: Request,
    schedule_id: str = Path(..., min_length=1, max_length=256, pattern=_SCHEDULE_ID_PATTERN),
) -> ScheduleDetailInfo | JSONResponse:
    """Get details of a specific schedule."""
    gw: Gateway = request.app
    detail = await gw.get_schedule(schedule_id)
    if detail is None:
        return not_found("schedule", schedule_id)

    return ScheduleDetailInfo(**detail)


@router.delete(
    "/schedules/{schedule_id:path}",
    summary="Delete admin schedule",
    description="Delete an admin-created schedule. Workspace schedules cannot be deleted.",
    tags=["Schedules"],
    responses={
        200: {"description": "Schedule deleted."},
        400: {"description": "Cannot delete a workspace schedule."},
        **build_responses(auth=True, not_found=True),
    },
    dependencies=[Depends(RequireScope("schedules:manage"))],
)
async def delete_schedule(
    request: Request,
    schedule_id: str = Path(..., min_length=1, max_length=256, pattern=_SCHEDULE_ID_PATTERN),
) -> JSONResponse:
    """Delete an admin-created schedule."""
    gw: Gateway = request.app
    if gw.scheduler is None:
        return error_response(404, "scheduler_not_active", "Scheduler is not active")

    # Check if schedule exists and its source
    detail = await gw.get_schedule(schedule_id)
    if detail is None:
        return not_found("schedule", schedule_id)

    if detail.get("source") != "admin":
        return error_response(
            400,
            "cannot_delete_workspace_schedule",
            "Workspace schedules cannot be deleted. Remove them from AGENT.md instead.",
        )

    ok = await gw.delete_admin_schedule(schedule_id)
    if not ok:
        return not_found("schedule", schedule_id)

    return JSONResponse(
        status_code=200,
        content={"status": "deleted", "schedule_id": schedule_id},
    )


@router.post(
    "/schedules/{schedule_id:path}/pause",
    summary="Pause schedule",
    description="Pause a schedule to stop future cron fires.",
    tags=["Schedules"],
    responses=build_responses(auth=True, not_found=True),
    dependencies=[Depends(RequireScope("schedules:manage"))],
)
async def pause_schedule(
    request: Request,
    schedule_id: str = Path(..., min_length=1, max_length=256, pattern=_SCHEDULE_ID_PATTERN),
) -> JSONResponse:
    """Pause a schedule (stops future cron fires)."""
    gw: Gateway = request.app
    if gw.scheduler is None:
        return error_response(404, "scheduler_not_active", "Scheduler is not active")

    ok = await gw.pause_schedule(schedule_id)
    if not ok:
        return not_found("schedule", schedule_id)

    return JSONResponse(
        status_code=200,
        content={"status": "paused", "schedule_id": schedule_id},
    )


@router.post(
    "/schedules/{schedule_id:path}/resume",
    summary="Resume schedule",
    description="Resume a previously paused schedule.",
    tags=["Schedules"],
    responses=build_responses(auth=True, not_found=True),
    dependencies=[Depends(RequireScope("schedules:manage"))],
)
async def resume_schedule(
    request: Request,
    schedule_id: str = Path(..., min_length=1, max_length=256, pattern=_SCHEDULE_ID_PATTERN),
) -> JSONResponse:
    """Resume a paused schedule."""
    gw: Gateway = request.app
    if gw.scheduler is None:
        return error_response(404, "scheduler_not_active", "Scheduler is not active")

    ok = await gw.resume_schedule(schedule_id)
    if not ok:
        return not_found("schedule", schedule_id)

    return JSONResponse(
        status_code=200,
        content={"status": "resumed", "schedule_id": schedule_id},
    )


@router.post(
    "/schedules/{schedule_id:path}/trigger",
    summary="Trigger schedule",
    description="Manually trigger a scheduled job to run immediately.",
    tags=["Schedules"],
    responses={
        202: {"description": "Accepted — job queued. Returns execution_id for polling."},
        **build_responses(auth=True, not_found=True),
    },
    dependencies=[Depends(RequireScope("schedules:manage"))],
)
async def trigger_schedule(
    request: Request,
    schedule_id: str = Path(..., min_length=1, max_length=256, pattern=_SCHEDULE_ID_PATTERN),
) -> JSONResponse:
    """Manually trigger a scheduled job (runs immediately)."""
    gw: Gateway = request.app
    if gw.scheduler is None:
        return error_response(404, "scheduler_not_active", "Scheduler is not active")

    execution_id = await gw.trigger_schedule(schedule_id)
    if execution_id is None:
        return not_found("schedule", schedule_id)

    return JSONResponse(
        status_code=202,
        content={
            "status": "triggered",
            "schedule_id": schedule_id,
            "execution_id": execution_id,
        },
    )



"""Tests for UserScheduleRecord domain model and UserScheduleRepository."""

from __future__ import annotations

from datetime import UTC, datetime

import pytest

from agent_gateway.persistence.domain import UserScheduleRecord
from agent_gateway.persistence.null import NullUserScheduleRepository


class TestUserScheduleRecordModel:
    """Test the UserScheduleRecord dataclass."""

    def test_defaults(self) -> None:
        record = UserScheduleRecord(
            id="sched-1",
            user_id="user-1",
            agent_id="agent-1",
            name="morning-check",
            cron_expr="0 8 * * 1-5",
            message="Check inbox",
        )
        assert record.enabled is True
        assert record.timezone == "UTC"
        assert record.input is None
        assert record.notify is None
        assert record.last_run_at is None

    def test_with_all_fields(self) -> None:
        now = datetime.now(UTC)
        record = UserScheduleRecord(
            id="sched-1",
            user_id="user-1",
            agent_id="agent-1",
            name="morning-check",
            cron_expr="0 8 * * 1-5",
            message="Check inbox",
            input={"folder": "INBOX"},
            enabled=False,
            timezone="US/Eastern",
            notify={"webhook": "https://example.com/hook"},
            last_run_at=now,
            next_run_at=now,
            created_at=now,
        )
        assert record.enabled is False
        assert record.timezone == "US/Eastern"
        assert record.notify is not None


class TestNullUserScheduleRepository:
    """Test the null implementation."""

    @pytest.fixture
    def repo(self) -> NullUserScheduleRepository:
        return NullUserScheduleRepository()

    async def test_create_is_noop(self, repo: NullUserScheduleRepository) -> None:
        record = UserScheduleRecord(
            id="s-1",
            user_id="u-1",
            agent_id="a-1",
            name="test",
            cron_expr="* * * * *",
            message="test",
        )
        await repo.create(record)  # should not raise

    async def test_get_returns_none(self, repo: NullUserScheduleRepository) -> None:
        result = await repo.get("s-1")
        assert result is None

    async def test_list_by_user_returns_empty(self, repo: NullUserScheduleRepository) -> None:
        assert await repo.list_by_user("u-1") == []

    async def test_delete_returns_false(self, repo: NullUserScheduleRepository) -> None:
        assert await repo.delete("s-1") is False

    async def test_update_enabled_is_noop(self, repo: NullUserScheduleRepository) -> None:
        await repo.update_enabled("s-1", False)  # should not raise

    async def test_update_last_run_is_noop(self, repo: NullUserScheduleRepository) -> None:
        now = datetime.now(UTC)
        await repo.update_last_run("s-1", now, now)  # should not raise


class TestUserScheduleSqlRepository:
    """Test SQL repository with in-memory SQLite."""

    @pytest.fixture
    async def repo(self):
        """Create an in-memory SQLite backend and return the user schedule repo."""
        from agent_gateway.persistence.backends.sqlite import SqliteBackend

        backend = SqliteBackend(path=":memory:")
        await backend.initialize()
        yield backend.user_schedule_repo
        await backend.dispose()

    async def test_create_and_get(self, repo) -> None:
        now = datetime.now(UTC)
        record = UserScheduleRecord(
            id="sched-1",
            user_id="user-1",
            agent_id="agent-1",
            name="morning-check",
            cron_expr="0 8 * * 1-5",
            message="Check inbox",
            input={"folder": "INBOX"},
            enabled=True,
            timezone="US/Eastern",
            created_at=now,
        )
        await repo.create(record)
        result = await repo.get("sched-1")
        assert result is not None
        assert result.user_id == "user-1"
        assert result.agent_id == "agent-1"
        assert result.name == "morning-check"
        assert result.cron_expr == "0 8 * * 1-5"

    async def test_list_by_user(self, repo) -> None:
        now = datetime.now(UTC)
        for i in range(3):
            await repo.create(
                UserScheduleRecord(
                    id=f"sched-{i}",
                    user_id="user-1",
                    agent_id="agent-1",
                    name=f"schedule-{i}",
                    cron_expr="* * * * *",
                    message="test",
                    created_at=now,
                )
            )
        # Different user
        await repo.create(
            UserScheduleRecord(
                id="sched-other",
                user_id="user-2",
                agent_id="agent-1",
                name="other",
                cron_expr="* * * * *",
                message="test",
                created_at=now,
            )
        )
        results = await repo.list_by_user("user-1")
        assert len(results) == 3

    async def test_update_enabled(self, repo) -> None:
        now = datetime.now(UTC)
        await repo.create(
            UserScheduleRecord(
                id="sched-1",
                user_id="user-1",
                agent_id="agent-1",
                name="test",
                cron_expr="* * * * *",
                message="test",
                enabled=True,
                created_at=now,
            )
        )
        await repo.update_enabled("sched-1", False)
        result = await repo.get("sched-1")
        assert result is not None
        assert result.enabled is False

    async def test_update_enabled_nonexistent(self, repo) -> None:
        # Should not raise for nonexistent schedule
        await repo.update_enabled("nonexistent", False)

    async def test_update_last_run(self, repo) -> None:
        now = datetime.now(UTC)
        await repo.create(
            UserScheduleRecord(
                id="sched-1",
                user_id="user-1",
                agent_id="agent-1",
                name="test",
                cron_expr="* * * * *",
                message="test",
                created_at=now,
            )
        )
        await repo.update_last_run("sched-1", now, now)
        result = await repo.get("sched-1")
        assert result is not None
        assert result.last_run_at is not None

    async def test_update_last_run_nonexistent(self, repo) -> None:
        now = datetime.now(UTC)
        await repo.update_last_run("nonexistent", now, now)

    async def test_delete(self, repo) -> None:
        now = datetime.now(UTC)
        await repo.create(
            UserScheduleRecord(
                id="sched-1",
                user_id="user-1",
                agent_id="agent-1",
                name="test",
                cron_expr="* * * * *",
                message="test",
                created_at=now,
            )
        )
        assert await repo.delete("sched-1") is True
        assert await repo.get("sched-1") is None
        assert await repo.delete("sched-1") is False



"""E2E tests for the travel-planner agent (multi-tool invocation)."""

from __future__ import annotations

import pytest
from httpx import AsyncClient

pytestmark = pytest.mark.e2e


async def test_invoke_travel_planner(client: AsyncClient) -> None:
    """Invoke travel-planner and verify multiple tools are called."""
    resp = await client.post(
        "/v1/agents/travel-planner/invoke",
        json={
            "message": "Plan the trip. Use all available tools.",
            "input": {
                "destination": "Paris",
                "origin": "London",
                "departure_date": "2025-07-15",
                "nights": 3,
            },
        },
        timeout=60,
    )
    assert resp.status_code == 200
    data = resp.json()
    assert data["status"] == "completed"
    assert data["agent_id"] == "travel-planner"
    # The agent should call at least 3 of the 4 travel tools
    assert data["usage"]["tool_calls"] >= 3


async def test_travel_planner_details(client: AsyncClient) -> None:
    """GET /v1/agents/travel-planner returns agent with 4 travel tools."""
    resp = await client.get("/v1/agents/travel-planner")
    assert resp.status_code == 200
    data = resp.json()
    assert data["id"] == "travel-planner"
    expected_tools = {"get-weather", "search-flights", "search-hotels", "search-activities"}
    assert expected_tools.issubset(set(data.get("tools", [])))



"""Unit tests for admin schedule API routes (POST /v1/schedules, DELETE /v1/schedules/{id})."""

from __future__ import annotations

from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

import pytest
from httpx import ASGITransport, AsyncClient

from agent_gateway.exceptions import ScheduleConflictError, ScheduleValidationError
from agent_gateway.gateway import Gateway


def _write_workspace(tmp_path: Path) -> Path:
    """Create a minimal workspace with one agent and persistence disabled."""
    (tmp_path / "gateway.yaml").write_text("persistence:\n  enabled: false\n")
    agents = tmp_path / "agents"
    agents.mkdir()
    (tmp_path / "skills").mkdir()
    (tmp_path / "tools").mkdir()

    agent_dir = agents / "test-agent"
    agent_dir.mkdir()
    (agent_dir / "AGENT.md").write_text("# Test Agent\n\nYou help with testing.\n")
    return tmp_path


@pytest.fixture
def workspace(tmp_path: Path) -> Path:
    return _write_workspace(tmp_path)


def _mock_scheduler() -> MagicMock:
    """Return a non-None mock to satisfy the `gw.scheduler is not None` guard.

    Uses AsyncMock for stop() so that gateway shutdown doesn't raise.
    """
    mock = MagicMock()
    mock.stop = AsyncMock()
    return mock


# ---------------------------------------------------------------------------
# POST /v1/schedules — create admin schedule
# ---------------------------------------------------------------------------

_VALID_PAYLOAD = {
    "agent_id": "test-agent",
    "name": "daily-report",
    "cron_expr": "0 9 * * *",
    "message": "Generate the daily report",
}


async def test_create_schedule_201(workspace: Path) -> None:
    """Happy path: valid payload returns 201 with schedule_id."""
    gw = Gateway(workspace=str(workspace), auth=False)
    async with gw:
        gw._scheduler = _mock_scheduler()
        gw.create_admin_schedule = AsyncMock(return_value="admin:test-agent:daily-report")  # type: ignore[method-assign]

        async with AsyncClient(
            transport=ASGITransport(app=gw),
            base_url="http://test",  # type: ignore[arg-type]
        ) as ac:
            resp = await ac.post("/v1/schedules", json=_VALID_PAYLOAD)

    assert resp.status_code == 201
    body = resp.json()
    assert body["status"] == "created"
    assert body["schedule_id"] == "admin:test-agent:daily-report"


async def test_create_schedule_400_bad_cron(workspace: Path) -> None:
    """ScheduleValidationError from gateway returns 400."""
    gw = Gateway(workspace=str(workspace), auth=False)
    async with gw:
        gw._scheduler = _mock_scheduler()
        gw.create_admin_schedule = AsyncMock(  # type: ignore[method-assign]
            side_effect=ScheduleValidationError("Invalid cron expression")
        )

        async with AsyncClient(
            transport=ASGITransport(app=gw),
            base_url="http://test",  # type: ignore[arg-type]
        ) as ac:
            resp = await ac.post("/v1/schedules", json=_VALID_PAYLOAD)

    assert resp.status_code == 400
    body = resp.json()
    assert body["error"]["code"] == "schedule_validation_error"


async def test_create_schedule_409_duplicate(workspace: Path) -> None:
    """ScheduleConflictError from gateway returns 409."""
    gw = Gateway(workspace=str(workspace), auth=False)
    async with gw:
        gw._scheduler = _mock_scheduler()
        gw.create_admin_schedule = AsyncMock(  # type: ignore[method-assign]
            side_effect=ScheduleConflictError("Schedule already exists")
        )

        async with AsyncClient(
            transport=ASGITransport(app=gw),
            base_url="http://test",  # type: ignore[arg-type]
        ) as ac:
            resp = await ac.post("/v1/schedules", json=_VALID_PAYLOAD)

    assert resp.status_code == 409
    body = resp.json()
    assert body["error"]["code"] == "schedule_conflict"


async def test_create_schedule_404_unknown_agent(workspace: Path) -> None:
    """Unknown agent_id returns 404."""
    gw = Gateway(workspace=str(workspace), auth=False)
    async with gw:
        gw._scheduler = _mock_scheduler()

        payload = {**_VALID_PAYLOAD, "agent_id": "nonexistent"}
        async with AsyncClient(
            transport=ASGITransport(app=gw),
            base_url="http://test",  # type: ignore[arg-type]
        ) as ac:
            resp = await ac.post("/v1/schedules", json=payload)

    assert resp.status_code == 404
    body = resp.json()
    assert body["error"]["code"] == "agent_not_found"


# ---------------------------------------------------------------------------
# DELETE /v1/schedules/{id} — delete admin schedule
# ---------------------------------------------------------------------------


async def test_delete_admin_schedule_200(workspace: Path) -> None:
    """Happy path: delete an admin schedule returns 200."""
    gw = Gateway(workspace=str(workspace), auth=False)
    async with gw:
        gw._scheduler = _mock_scheduler()
        gw.get_schedule = AsyncMock(  # type: ignore[method-assign]
            return_value={"source": "admin", "schedule_id": "admin:test-agent:daily"}
        )
        gw.delete_admin_schedule = AsyncMock(return_value=True)  # type: ignore[method-assign]

        async with AsyncClient(
            transport=ASGITransport(app=gw),
            base_url="http://test",  # type: ignore[arg-type]
        ) as ac:
            resp = await ac.delete("/v1/schedules/admin:test-agent:daily")

    assert resp.status_code == 200
    body = resp.json()
    assert body["status"] == "deleted"
    assert body["schedule_id"] == "admin:test-agent:daily"


async def test_delete_workspace_schedule_400(workspace: Path) -> None:
    """Attempting to delete a workspace schedule returns 400."""
    gw = Gateway(workspace=str(workspace), auth=False)
    async with gw:
        gw._scheduler = _mock_scheduler()
        gw.get_schedule = AsyncMock(  # type: ignore[method-assign]
            return_value={"source": "workspace", "schedule_id": "test-agent:daily"}
        )

        async with AsyncClient(
            transport=ASGITransport(app=gw),
            base_url="http://test",  # type: ignore[arg-type]
        ) as ac:
            resp = await ac.delete("/v1/schedules/test-agent:daily")

    assert resp.status_code == 400
    body = resp.json()
    assert body["error"]["code"] == "cannot_delete_workspace_schedule"



"""In-process asyncio.Queue backend — development and testing only.

Limitations:
- Jobs are lost on server restart (no persistence).
- Single-process only — cannot be shared across workers.
- ``--worker-only`` mode is not supported with this backend.
"""

from __future__ import annotations

import asyncio

from agent_gateway.queue.models import ExecutionJob


class MemoryQueue:
    """Wraps ``asyncio.Queue`` to implement the ``ExecutionQueue`` protocol."""

    def __init__(self) -> None:
        self._queue: asyncio.Queue[ExecutionJob] = asyncio.Queue()
        self._cancelled: set[str] = set()

    async def initialize(self) -> None:
        pass

    async def dispose(self) -> None:
        self._cancelled.clear()

    async def enqueue(self, job: ExecutionJob) -> None:
        await self._queue.put(job)

    async def dequeue(self, timeout: float = 0) -> ExecutionJob | None:  # noqa: ASYNC109
        try:
            if timeout <= 0:
                return self._queue.get_nowait()
            return await asyncio.wait_for(self._queue.get(), timeout=timeout)
        except (asyncio.QueueEmpty, TimeoutError):
            return None

    async def ack(self, job_id: str) -> None:
        self._cancelled.discard(job_id)

    async def nack(self, job_id: str) -> None:
        # For memory backend, we cannot re-enqueue without the original job.
        # The worker must handle retry logic by re-enqueuing explicitly.
        pass

    async def request_cancel(self, job_id: str) -> bool:
        self._cancelled.add(job_id)
        return True

    async def is_cancelled(self, job_id: str) -> bool:
        return job_id in self._cancelled

    async def length(self) -> int:
        return self._queue.qsize()



"""Generate JSON Schema from Python function signatures for LLM tool declarations."""

from __future__ import annotations

import inspect
from typing import Annotated, Any, Literal, Union, get_args, get_origin, get_type_hints

from pydantic import BaseModel

# Python type -> JSON Schema type mapping
TYPE_MAP: dict[type, str] = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    list: "array",
    dict: "object",
}


def schema_from_function(fn: Any) -> dict[str, Any]:
    """Generate JSON Schema parameters from a function's type hints.

    Supports 4 modes:
    1. Single Pydantic model parameter -> model.model_json_schema()
    2. Annotated[type, "description"] parameters -> type + description extracted
    3. Bare type hints -> type inferred, parameter name used as description
    4. No hints -> all params treated as string
    """
    try:
        hints = get_type_hints(fn, include_extras=True)
    except Exception:
        # Fallback: use raw annotations when get_type_hints fails
        # (e.g., locally-defined classes with `from __future__ import annotations`)
        hints = _get_raw_annotations(fn)
    sig = inspect.signature(fn)
    properties: dict[str, Any] = {}
    required: list[str] = []

    for name, param in sig.parameters.items():
        if name in ("self", "cls", "return"):
            continue

        hint = hints.get(name, str)

        # Check if single Pydantic model parameter
        if inspect.isclass(hint) and issubclass(hint, BaseModel):
            return hint.model_json_schema()

        prop = _hint_to_schema(hint, name)
        properties[name] = prop

        # Required if no default
        if param.default is inspect.Parameter.empty:
            required.append(name)

    schema: dict[str, Any] = {"type": "object", "properties": properties}
    if required:
        schema["required"] = required
    return schema


def _hint_to_schema(hint: Any, name: str) -> dict[str, Any]:
    """Convert a single type hint to a JSON Schema property."""
    origin = get_origin(hint)
    args = get_args(hint)

    # Annotated[type, "description"]
    if origin is Annotated:
        base_type = args[0]
        description = next((a for a in args[1:] if isinstance(a, str)), name)
        prop = _base_type_to_schema(base_type)
        prop["description"] = description
        return prop

    # Optional[X] is Union[X, None]
    if origin is Union:
        non_none = [a for a in args if a is not type(None)]
        if len(non_none) == 1:
            prop = _base_type_to_schema(non_none[0])
            prop["description"] = name
            return prop

    # Literal["A", "B"] -> enum
    if origin is Literal:
        return {"type": "string", "enum": list(args), "description": name}

    # list[X]
    if origin is list and args:
        item_schema = _base_type_to_schema(args[0])
        return {"type": "array", "items": item_schema, "description": name}

    # Plain type
    prop = _base_type_to_schema(hint)
    prop["description"] = name
    return prop


def _base_type_to_schema(t: Any) -> dict[str, Any]:
    """Map a Python type to JSON Schema."""
    json_type = TYPE_MAP.get(t, "string")
    return {"type": json_type}


def _get_raw_annotations(fn: Any) -> dict[str, Any]:
    """Extract raw annotations, evaluating string annotations against the function's globals."""
    raw = getattr(fn, "__annotations__", {})
    globalns = getattr(fn, "__globals__", {})
    hints: dict[str, Any] = {}
    for name, annotation in raw.items():
        if isinstance(annotation, str):
            try:
                hints[name] = eval(annotation, globalns)  # noqa: S307
            except Exception:
                hints[name] = str
        else:
            hints[name] = annotation
    return hints



"""Tests for schedule and notification parsing in AgentDefinition.load()."""

from __future__ import annotations

from pathlib import Path

from agent_gateway.workspace.agent import AgentDefinition


def _write_agent(tmp_path: Path, frontmatter: str, body: str = "# Agent\n\nDo stuff.\n") -> Path:
    agents = tmp_path / "agents"
    agents.mkdir(exist_ok=True)
    agent_dir = agents / "test-agent"
    agent_dir.mkdir(exist_ok=True)
    (agent_dir / "AGENT.md").write_text(f"---\n{frontmatter}---\n\n{body}")
    return agent_dir


def test_valid_schedule_parsed(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'schedules:\n  - name: daily\n    cron: "0 9 * * *"\n    message: "Run"\n',
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.schedules) == 1
    assert agent.schedules[0].name == "daily"
    assert agent.schedules[0].cron == "0 9 * * *"


def test_schedule_with_timezone(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'schedules:\n  - name: daily\n    cron: "0 9 * * *"\n'
        '    message: "Run"\n    timezone: "Europe/London"\n',
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert agent.schedules[0].timezone == "Europe/London"


def test_schedule_invalid_timezone_skipped(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'schedules:\n  - name: daily\n    cron: "0 9 * * *"\n'
        '    message: "Run"\n    timezone: "Invalid/Nowhere"\n',
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.schedules) == 0


def test_schedule_invalid_cron_skipped(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'schedules:\n  - name: daily\n    cron: "bad cron"\n    message: "Run"\n',
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.schedules) == 0


def test_schedule_missing_required_field_skipped(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'schedules:\n  - name: daily\n    cron: "0 9 * * *"\n',  # missing message
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.schedules) == 0


def test_schedule_non_dict_entry_skipped(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        "schedules:\n  - not-a-dict\n",
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.schedules) == 0


def test_schedule_duplicate_name_skipped(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'schedules:\n  - name: daily\n    cron: "0 9 * * *"\n'
        '    message: "First"\n'
        '  - name: daily\n    cron: "0 10 * * *"\n'
        '    message: "Dupe"\n',
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.schedules) == 1
    assert agent.schedules[0].message == "First"


def test_notification_config_parsed(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'notifications:\n  on_complete:\n    - channel: log\n      target: ""\n',
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.notifications.on_complete) == 1
    assert agent.notifications.on_complete[0].channel == "log"


def test_notification_invalid_target_skipped(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        "notifications:\n  on_complete:\n    - not-a-dict\n",
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.notifications.on_complete) == 0


def test_notification_missing_channel_skipped(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'notifications:\n  on_complete:\n    - target: "test"\n',  # missing channel
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.notifications.on_complete) == 0


def test_notification_non_dict_raw(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        "notifications: not-a-dict\n",
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.notifications.on_complete) == 0


def test_notification_non_list_targets(tmp_path: Path) -> None:
    agent_dir = _write_agent(
        tmp_path,
        'notifications:\n  on_complete: "not-a-list"\n',
    )
    agent = AgentDefinition.load(agent_dir)
    assert agent is not None
    assert len(agent.notifications.on_complete) == 0



"""Tests for execution engine error isolation."""

from __future__ import annotations

from typing import Any

import pytest

from agent_gateway.engine.models import StopReason, ToolContext
from agent_gateway.workspace.registry import ResolvedTool
from tests.test_engine.conftest import (
    make_agent,
    make_engine,
    make_llm_response,
    make_resolved_tool,
    make_skill,
    make_tool_call,
    make_workspace,
    simple_tool_executor,
)


class TestUnknownTool:
    @pytest.mark.asyncio
    async def test_unknown_tool_returns_error_to_llm(self) -> None:
        """LLM calls a tool that doesn't exist → error in tool result, loop continues."""
        engine, _, _ = make_engine(
            responses=[
                make_llm_response(
                    tool_calls=[make_tool_call(name="nonexistent", arguments={}, call_id="c1")]
                ),
                make_llm_response(text="OK, that tool doesn't exist"),
            ],
        )
        agent = make_agent()
        workspace = make_workspace()

        result = await engine.execute(
            agent, "call nonexistent", workspace, tool_executor=simple_tool_executor
        )

        assert result.stop_reason == StopReason.COMPLETED
        assert result.raw_text == "OK, that tool doesn't exist"


class TestToolPermissionDenied:
    @pytest.mark.asyncio
    async def test_tool_not_permitted_for_agent(self) -> None:
        """Tool exists but is restricted → error returned, loop continues."""
        restricted_tool = make_resolved_tool(name="restricted", allowed_agents=["other-agent"])
        engine, _, _ = make_engine(
            responses=[
                make_llm_response(
                    tool_calls=[make_tool_call(name="restricted", arguments={}, call_id="c1")]
                ),
                make_llm_response(text="Permission denied, moving on"),
            ],
            tools=[restricted_tool],
        )
        # The tool is registered but agent doesn't have access via resolve_for_agent
        # In practice, resolve_for_agent filters it out, so it appears as unknown
        skill = make_skill(tools=["restricted"])
        agent = make_agent(skills=["test-skill"])
        workspace = make_workspace(agents={"test-agent": agent}, skills={"test-skill": skill})

        result = await engine.execute(
            agent, "call restricted", workspace, tool_executor=simple_tool_executor
        )

        assert result.stop_reason == StopReason.COMPLETED


class TestToolException:
    @pytest.mark.asyncio
    async def test_tool_raises_exception(self) -> None:
        """Tool raises exception → error returned to LLM, loop continues."""
        echo_tool = make_resolved_tool(name="failing-tool")

        async def failing_executor(
            tool: ResolvedTool, arguments: dict[str, Any], context: ToolContext
        ) -> Any:
            raise ValueError("Tool crashed!")

        engine, _, _ = make_engine(
            responses=[
                make_llm_response(
                    tool_calls=[make_tool_call(name="failing-tool", arguments={}, call_id="c1")]
                ),
                make_llm_response(text="Tool failed, but I handled it"),
            ],
            tools=[echo_tool],
        )
        skill = make_skill(tools=["failing-tool"])
        agent = make_agent(skills=["test-skill"])
        workspace = make_workspace(agents={"test-agent": agent}, skills={"test-skill": skill})

        result = await engine.execute(agent, "test", workspace, tool_executor=failing_executor)

        assert result.stop_reason == StopReason.COMPLETED
        assert result.raw_text == "Tool failed, but I handled it"


class TestOversizedResult:
    @pytest.mark.asyncio
    async def test_tool_result_truncated(self) -> None:
        """Tool result exceeding 32KB is truncated."""
        echo_tool = make_resolved_tool(name="big-tool")

        large_output = "x" * 40_000

        async def big_executor(
            tool: ResolvedTool, arguments: dict[str, Any], context: ToolContext
        ) -> Any:
            return large_output

        engine, mock_llm, _ = make_engine(
            responses=[
                make_llm_response(
                    tool_calls=[make_tool_call(name="big-tool", arguments={}, call_id="c1")]
                ),
                make_llm_response(text="Got truncated result"),
            ],
            tools=[echo_tool],
        )
        skill = make_skill(tools=["big-tool"])
        agent = make_agent(skills=["test-skill"])
        workspace = make_workspace(agents={"test-agent": agent}, skills={"test-skill": skill})

        result = await engine.execute(agent, "test", workspace, tool_executor=big_executor)

        assert result.stop_reason == StopReason.COMPLETED
        # Verify the tool result message was truncated
        tool_msg = mock_llm.calls[1]["messages"][-1]  # Last message before 2nd LLM call
        assert "[truncated: result exceeded 32KB limit]" in tool_msg["content"]


class TestInvalidToolArguments:
    @pytest.mark.asyncio
    async def test_invalid_arguments_returns_error_to_llm(self) -> None:
        """Tool called with invalid arguments → validation error to LLM, loop continues."""
        from agent_gateway.workspace.registry import CodeTool

        code_tool = CodeTool(
            name="strict-tool",
            description="A tool with strict schema",
            fn=lambda: None,
            parameters_schema={
                "type": "object",
                "properties": {"count": {"type": "integer"}},
                "required": ["count"],
            },
        )
        strict_tool = ResolvedTool(
            name="strict-tool",
            description="A tool with strict schema",
            source="code",
            llm_declaration=code_tool.to_llm_declaration(),
            parameters_schema=code_tool.parameters_schema,
            code_tool=code_tool,
        )

        engine, mock_llm, _ = make_engine(
            responses=[
                make_llm_response(
                    tool_calls=[
                        make_tool_call(
                            name="strict-tool",
                            arguments={"count": "not-a-number"},
                            call_id="c1",
                        )
                    ]
                ),
                make_llm_response(text="Fixed the arguments"),
            ],
            tools=[strict_tool],
        )
        skill = make_skill(tools=["strict-tool"])
        agent = make_agent(skills=["test-skill"])
        workspace = make_workspace(agents={"test-agent": agent}, skills={"test-skill": skill})

        result = await engine.execute(agent, "test", workspace, tool_executor=simple_tool_executor)

        assert result.stop_reason == StopReason.COMPLETED
        assert result.raw_text == "Fixed the arguments"
        # Verify the error was sent back to the LLM
        tool_msg = mock_llm.calls[1]["messages"][-1]
        assert "Invalid arguments" in tool_msg["content"]


class TestSanitizedErrorMessages:
    @pytest.mark.asyncio
    async def test_tool_exception_does_not_leak_details(self) -> None:
        """Tool exception error message does not expose internal details."""
        echo_tool = make_resolved_tool(name="leaky-tool")

        async def leaky_executor(
            tool: ResolvedTool, arguments: dict[str, Any], context: ToolContext
        ) -> Any:
            raise ValueError("secret connection string: postgres://user:pass@host/db")

        engine, mock_llm, _ = make_engine(
            responses=[
                make_llm_response(
                    tool_calls=[make_tool_call(name="leaky-tool", arguments={}, call_id="c1")]
                ),
                make_llm_response(text="Handled"),
            ],
            tools=[echo_tool],
        )
        skill = make_skill(tools=["leaky-tool"])
        agent = make_agent(skills=["test-skill"])
        workspace = make_workspace(agents={"test-agent": agent}, skills={"test-skill": skill})

        result = await engine.execute(agent, "test", workspace, tool_executor=leaky_executor)

        assert result.stop_reason == StopReason.COMPLETED
        tool_msg = mock_llm.calls[1]["messages"][-1]
        assert "postgres://" not in tool_msg["content"]
        assert "failed unexpectedly" in tool_msg["content"]


class TestNoToolExecutor:
    @pytest.mark.asyncio
    async def test_no_executor_configured(self) -> None:
        """When no tool executor is provided, tools return an error."""
        echo_tool = make_resolved_tool(name="echo")
        engine, _, _ = make_engine(
            responses=[
                make_llm_response(
                    tool_calls=[make_tool_call(name="echo", arguments={}, call_id="c1")]
                ),
                make_llm_response(text="No executor"),
            ],
            tools=[echo_tool],
        )
        skill = make_skill(tools=["echo"])
        agent = make_agent(skills=["test-skill"])
        workspace = make_workspace(agents={"test-agent": agent}, skills={"test-skill": skill})

        # No tool_executor passed
        result = await engine.execute(agent, "test", workspace)

        assert result.stop_reason == StopReason.COMPLETED



"""Integration tests for dashboard routes (router.py).

Uses a real Gateway with the dashboard enabled and httpx.AsyncClient
as the test transport, so the full Starlette middleware stack is exercised.
"""

from __future__ import annotations

from pathlib import Path
from unittest.mock import AsyncMock

from httpx import ASGITransport, AsyncClient

from agent_gateway.gateway import Gateway

FIXTURE_WORKSPACE = Path(__file__).resolve().parent.parent / "fixtures" / "workspace"


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _mock_repos(gw: Gateway) -> None:
    """Replace persistence repos with async mocks returning empty/zero defaults."""

    exec_repo = AsyncMock()
    exec_repo.list_all.return_value = []
    exec_repo.count_all.return_value = 0
    exec_repo.get.return_value = None
    exec_repo.get_with_steps.return_value = None
    exec_repo.list_conversations_summary.return_value = []
    exec_repo.count_conversations.return_value = 0
    exec_repo.get_summary_stats.return_value = {
        "total_executions": 0,
        "total_cost_usd": 0.0,
        "success_count": 0,
        "avg_duration_ms": 0.0,
    }
    exec_repo.cost_by_day.return_value = []
    exec_repo.executions_by_day.return_value = []
    exec_repo.cost_by_agent.return_value = []
    exec_repo.list_by_session.return_value = []
    exec_repo.get_schedule_stats.return_value = {
        "total_scheduled": 0,
        "active_schedules": 0,
        "success": 0,
        "failed": 0,
        "running": 0,
    }
    exec_repo.list_children.return_value = []
    exec_repo.cost_by_root_execution.return_value = 0.0

    schedule_repo = AsyncMock()
    schedule_repo.list_all.return_value = []
    schedule_repo.get.return_value = None

    user_agent_config_repo = AsyncMock()
    user_agent_config_repo.list_by_user.return_value = []
    user_agent_config_repo.get.return_value = None

    user_schedule_repo = AsyncMock()
    user_schedule_repo.list_by_user.return_value = []

    gw._execution_repo = exec_repo  # type: ignore[assignment]
    gw._schedule_repo = schedule_repo  # type: ignore[assignment]
    gw._user_agent_config_repo = user_agent_config_repo  # type: ignore[assignment]
    gw._user_schedule_repo = user_schedule_repo  # type: ignore[assignment]


async def _make_client(
    gw: Gateway,
) -> AsyncClient:
    """Build an httpx AsyncClient wired to the gateway."""
    transport = ASGITransport(app=gw)
    return AsyncClient(transport=transport, base_url="http://test")


async def _login(
    client: AsyncClient,
    username: str = "testuser",
    password: str = "testpass",
) -> None:
    """POST valid credentials so the session cookie is set."""
    await client.post(
        "/dashboard/login",
        data={"username": username, "password": password},
        follow_redirects=False,
    )


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------


class TestLoginPage:
    async def test_get_login_returns_200(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                resp = await client.get("/dashboard/login")
                assert resp.status_code == 200
                assert "text/html" in resp.headers["content-type"]

    async def test_post_valid_creds_redirects(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                resp = await client.post(
                    "/dashboard/login",
                    data={"username": "testuser", "password": "testpass"},
                    follow_redirects=False,
                )
                assert resp.status_code == 303
                assert "/dashboard/" in resp.headers.get("location", "")

    async def test_post_invalid_creds_returns_401(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                resp = await client.post(
                    "/dashboard/login",
                    data={"username": "testuser", "password": "wrong"},
                    follow_redirects=False,
                )
                assert resp.status_code == 401


class TestLogout:
    async def test_logout_redirects_to_login(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client)
                resp = await client.post("/dashboard/logout", follow_redirects=False)
                assert resp.status_code == 303
                assert "/dashboard/login" in resp.headers.get("location", "")


class TestProtectedRoutes:
    async def test_unauthenticated_redirects(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                resp = await client.get("/dashboard/", follow_redirects=False)
                assert resp.status_code == 302
                assert "/dashboard/login" in resp.headers.get("location", "")


class TestAgentsPage:
    async def test_agents_page_returns_200(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client)
                resp = await client.get("/dashboard/")
                assert resp.status_code == 200
                assert "text/html" in resp.headers["content-type"]

    async def test_agents_htmx_partial(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client)
                resp = await client.get(
                    "/dashboard/agents",
                    headers={"HX-Request": "true"},
                )
                assert resp.status_code == 200


class TestExecutionsPage:
    async def test_executions_page_returns_200(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get("/dashboard/executions")
                assert resp.status_code == 200

    async def test_executions_with_filters(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get(
                    "/dashboard/executions",
                    params={"agent_id": "x", "status": "completed"},
                )
                assert resp.status_code == 200

    async def test_execution_detail_not_found(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get("/dashboard/executions/nonexistent")
                assert resp.status_code == 404


class TestConversationsPage:
    async def test_conversations_page_returns_200(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client)
                resp = await client.get("/dashboard/conversations")
                assert resp.status_code == 200

    async def test_conversation_detail_not_found(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client)
                resp = await client.get("/dashboard/conversations/nonexistent")
                assert resp.status_code == 404


class TestAnalyticsPage:
    async def test_analytics_page_returns_200(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get("/dashboard/analytics")
                assert resp.status_code == 200

    async def test_analytics_charts_htmx(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get(
                    "/dashboard/analytics",
                    headers={"HX-Request": "true"},
                )
                assert resp.status_code == 200


class TestSchedulesPage:
    async def test_schedules_page_returns_200(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get("/dashboard/schedules")
                assert resp.status_code == 200


class TestAdminRoutes:
    async def test_toggle_schedule_non_admin_403(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="testuser", password="testpass")
                resp = await client.post(
                    "/dashboard/schedules/some-id/toggle",
                    follow_redirects=False,
                )
                assert resp.status_code == 303

    async def test_toggle_schedule_admin_works(self) -> None:
        from agent_gateway.persistence.domain import ScheduleRecord

        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            # Make schedule_repo.get return a real record
            record = ScheduleRecord(
                id="sched-1",
                agent_id="test-agent",
                name="test schedule",
                cron_expr="*/5 * * * *",
                message="run",
                enabled=True,
            )
            schedule_repo = gw._schedule_repo
            assert isinstance(schedule_repo, AsyncMock)
            schedule_repo.get.return_value = record
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.post(
                    "/dashboard/schedules/sched-1/toggle",
                    follow_redirects=False,
                )
                assert resp.status_code == 303

    async def test_toggle_schedule_not_found(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.post(
                    "/dashboard/schedules/nonexistent/toggle",
                    follow_redirects=False,
                )
                assert resp.status_code == 404


class TestScheduleDetailPage:
    async def test_schedule_detail_returns_200_with_recent_runs(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            # Mock get_schedule to return a schedule dict
            schedule_dict = {
                "id": "sched-test",
                "agent_id": "test-agent",
                "name": "test schedule",
                "cron_expr": "*/5 * * * *",
                "message": "run",
                "instructions": "",
                "input": {},
                "enabled": True,
                "timezone": "UTC",
                "source": "workspace",
                "next_run_at": None,
                "last_run_at": None,
                "created_at": None,
            }
            gw.get_schedule = AsyncMock(return_value=schedule_dict)  # type: ignore[method-assign]
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get("/dashboard/schedules/sched-test/detail")
                assert resp.status_code == 200
                assert "text/html" in resp.headers["content-type"]
                # Verify list_all was called with schedule_id filter
                exec_repo = gw._execution_repo
                assert isinstance(exec_repo, AsyncMock)
                exec_repo.list_all.assert_called_with(schedule_id="sched-test", limit=20)

    async def test_schedule_detail_not_found(self) -> None:
        gw = Gateway(workspace=str(FIXTURE_WORKSPACE), auth=False, title="Test")
        gw.use_dashboard(
            auth_password="testpass",
            auth_username="testuser",
            admin_username="admin",
            admin_password="adminpass",
        )
        async with gw:
            _mock_repos(gw)
            gw.get_schedule = AsyncMock(return_value=None)  # type: ignore[method-assign]
            client = await _make_client(gw)
            async with client:
                await _login(client, username="admin", password="adminpass")
                resp = await client.get("/dashboard/schedules/nonexistent/detail")
                assert resp.status_code == 404



"""Tests for input context validation."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel

from agent_gateway.engine.input import (
    resolve_input_schema,
    validate_input,
    validate_input_pydantic,
)

SIMPLE_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "deal_id": {"type": "string"},
        "amount": {"type": "number"},
    },
    "required": ["deal_id"],
}


class DealInput(BaseModel):
    deal_id: str
    amount: float = 0.0
    tier: str = "standard"


# -- resolve_input_schema -----------------------------------------------------


class TestResolveInputSchema:
    def test_dict_passthrough(self) -> None:
        json_schema, model_cls = resolve_input_schema(SIMPLE_SCHEMA)
        assert json_schema is SIMPLE_SCHEMA
        assert model_cls is None

    def test_pydantic_model(self) -> None:
        json_schema, model_cls = resolve_input_schema(DealInput)
        assert model_cls is DealInput
        assert json_schema["type"] == "object"
        assert "deal_id" in json_schema["properties"]
        assert "amount" in json_schema["properties"]


# -- validate_input (JSON Schema) --------------------------------------------


class TestValidateInput:
    def test_valid_context(self) -> None:
        errors = validate_input({"deal_id": "D-123", "amount": 50000}, SIMPLE_SCHEMA)
        assert errors == []

    def test_valid_context_optional_missing(self) -> None:
        errors = validate_input({"deal_id": "D-123"}, SIMPLE_SCHEMA)
        assert errors == []

    def test_missing_required_field(self) -> None:
        errors = validate_input({"amount": 100}, SIMPLE_SCHEMA)
        assert len(errors) == 1
        assert "deal_id" in errors[0]

    def test_wrong_type(self) -> None:
        errors = validate_input({"deal_id": 123}, SIMPLE_SCHEMA)
        assert len(errors) == 1

    def test_empty_context_with_required(self) -> None:
        errors = validate_input({}, SIMPLE_SCHEMA)
        assert len(errors) == 1
        assert "deal_id" in errors[0]

    def test_none_context_with_required(self) -> None:
        errors = validate_input(None, SIMPLE_SCHEMA)
        assert len(errors) == 1
        assert "deal_id" in errors[0]

    def test_none_context_no_required(self) -> None:
        schema: dict[str, Any] = {
            "type": "object",
            "properties": {"note": {"type": "string"}},
        }
        errors = validate_input(None, schema)
        assert errors == []

    def test_empty_context_no_required(self) -> None:
        schema: dict[str, Any] = {
            "type": "object",
            "properties": {"note": {"type": "string"}},
        }
        errors = validate_input({}, schema)
        assert errors == []

    def test_extra_fields_allowed_by_default(self) -> None:
        errors = validate_input(
            {"deal_id": "D-1", "extra": "field"},
            SIMPLE_SCHEMA,
        )
        assert errors == []


# -- validate_input_pydantic --------------------------------------------------


class TestValidateInputPydantic:
    def test_valid_context(self) -> None:
        errors = validate_input_pydantic({"deal_id": "D-1", "amount": 100.0}, DealInput)
        assert errors == []

    def test_missing_required(self) -> None:
        errors = validate_input_pydantic({"amount": 100.0}, DealInput)
        assert len(errors) >= 1

    def test_wrong_type(self) -> None:
        errors = validate_input_pydantic({"deal_id": 123}, DealInput)
        assert len(errors) >= 1

    def test_none_context_with_required(self) -> None:
        errors = validate_input_pydantic(None, DealInput)
        assert len(errors) >= 1

    def test_empty_context_with_required(self) -> None:
        errors = validate_input_pydantic({}, DealInput)
        assert len(errors) >= 1

    def test_defaults_applied(self) -> None:
        errors = validate_input_pydantic({"deal_id": "D-1"}, DealInput)
        assert errors == []



"""LLM client — wraps LiteLLM for production use with failover and cost tracking."""

from __future__ import annotations

import json
import logging
from collections.abc import AsyncIterator
from dataclasses import dataclass, field
from typing import Any

import litellm
from litellm import ModelResponse
from litellm.router import (  # type: ignore[attr-defined]
    AllowedFailsPolicy,
    RetryPolicy,
    Router,
)

from agent_gateway.config import GatewayConfig, ModelConfig
from agent_gateway.engine.models import ToolCall
from agent_gateway.exceptions import ExecutionError
from agent_gateway.workspace.agent import AgentModelConfig

logger = logging.getLogger(__name__)

# Suppress LiteLLM's verbose output
litellm.suppress_debug_info = True


@dataclass
class LLMResponse:
    """Parsed response from an LLM call."""

    text: str | None = None
    tool_calls: list[ToolCall] = field(default_factory=list)
    model: str = ""
    input_tokens: int = 0
    output_tokens: int = 0
    cost: float = 0.0


def _build_model_list(config: ModelConfig) -> list[dict[str, Any]]:
    """Build LiteLLM Router model_list from gateway config."""
    models: list[dict[str, Any]] = [
        {
            "model_name": "default",
            "litellm_params": {"model": config.default},
        }
    ]
    if config.fallback:
        models.append(
            {
                "model_name": "fallback",
                "litellm_params": {"model": config.fallback},
            }
        )
    return models


def _parse_tool_calls(response: ModelResponse) -> list[ToolCall]:
    """Extract tool calls from an LLM response."""
    tool_calls: list[ToolCall] = []
    choices = response.choices
    if not choices:
        return tool_calls

    message = choices[0].message
    if not message or not message.tool_calls:
        return tool_calls

    for tc in message.tool_calls:
        args = tc.function.arguments
        if isinstance(args, str):
            try:
                args = json.loads(args)
            except (json.JSONDecodeError, TypeError):
                args = {"_raw": args}

        tool_calls.append(
            ToolCall(
                name=tc.function.name,
                arguments=args if isinstance(args, dict) else {"_raw": args},
                call_id=tc.id or "",
            )
        )
    return tool_calls


def _extract_cost(response: ModelResponse) -> float:
    """Extract cost from response, falling back to litellm.completion_cost."""
    try:
        return float(litellm.completion_cost(completion_response=response))
    except Exception:
        logger.debug("Could not extract cost from LLM response", exc_info=True)
        return 0.0


class LLMClient:
    """Production LLM client with failover and cost tracking."""

    def __init__(self, config: GatewayConfig) -> None:
        self._config = config
        model_list = _build_model_list(config.model)

        fallback_list = []
        if config.model.fallback:
            fallback_list = [{"default": ["fallback"]}]

        self._router = Router(
            model_list=model_list,
            retry_policy=RetryPolicy(
                TimeoutError=2,
                RateLimitError=3,
                ContentPolicyViolationError=0,
                AuthenticationError=0,
            ),
            allowed_fails_policy=AllowedFailsPolicy(
                RateLimitErrorAllowedFails=5,
                TimeoutErrorAllowedFails=3,
            ),
            cooldown_time=60.0,
            fallbacks=fallback_list,
            cache_responses=False,
            set_verbose=False,
        )

    async def completion(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None = None,
        model: str | None = None,
        temperature: float | None = None,
        max_tokens: int | None = None,
        response_format: dict[str, Any] | None = None,
    ) -> LLMResponse:
        """Make an LLM completion call with automatic failover."""
        kwargs: dict[str, Any] = {
            "model": model or "default",
            "messages": messages,
        }
        if tools:
            kwargs["tools"] = tools
        if temperature is not None:
            kwargs["temperature"] = temperature
        if max_tokens is not None:
            kwargs["max_tokens"] = max_tokens
        if response_format is not None:
            kwargs["response_format"] = response_format

        try:
            response: ModelResponse = await self._router.acompletion(**kwargs)
        except Exception as e:
            logger.error("LLM call failed: %s", e)
            raise ExecutionError(f"LLM call failed: {e}") from e

        return self._parse_response(response)

    async def stream_completion(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None = None,
        model: str | None = None,
        temperature: float | None = None,
        max_tokens: int | None = None,
    ) -> AsyncIterator[dict[str, Any]]:
        """Stream an LLM completion, yielding chunk events.

        Yields dicts with keys:
        - {"type": "token", "content": str}
        - {"type": "tool_call", "index": int, "name": str, "arguments": str, "call_id": str}
        - {"type": "usage", "input_tokens": int, "output_tokens": int, "cost": float, "model": str}
        """
        kwargs: dict[str, Any] = {
            "model": model or "default",
            "messages": messages,
            "stream": True,
            "stream_options": {"include_usage": True},
        }
        if tools:
            kwargs["tools"] = tools
        if temperature is not None:
            kwargs["temperature"] = temperature
        if max_tokens is not None:
            kwargs["max_tokens"] = max_tokens

        try:
            response = await self._router.acompletion(**kwargs)
        except Exception as e:
            logger.error("LLM streaming call failed: %s", e)
            raise ExecutionError(f"LLM streaming call failed: {e}") from e

        # Track tool call state for accumulation
        active_tool_calls: dict[int, dict[str, str]] = {}
        accumulated_model = ""
        accumulated_input_tokens = 0
        accumulated_output_tokens = 0

        async for chunk in response:
            # Extract model info
            if hasattr(chunk, "model") and chunk.model:
                accumulated_model = chunk.model

            # Extract usage from final chunk
            if hasattr(chunk, "usage") and chunk.usage:
                accumulated_input_tokens = getattr(chunk.usage, "prompt_tokens", 0) or 0
                accumulated_output_tokens = getattr(chunk.usage, "completion_tokens", 0) or 0

            choices = chunk.choices
            if not choices:
                continue

            delta = choices[0].delta
            if delta is None:
                continue

            # Text content
            if delta.content:
                yield {"type": "token", "content": delta.content}

            # Tool calls
            if delta.tool_calls:
                for tc_delta in delta.tool_calls:
                    idx = tc_delta.index if hasattr(tc_delta, "index") else 0

                    if idx not in active_tool_calls:
                        # New tool call
                        active_tool_calls[idx] = {
                            "name": "",
                            "arguments": "",
                            "call_id": "",
                        }

                    if tc_delta.id:
                        active_tool_calls[idx]["call_id"] = tc_delta.id

                    if tc_delta.function:
                        if tc_delta.function.name:
                            active_tool_calls[idx]["name"] = tc_delta.function.name
                        if tc_delta.function.arguments:
                            active_tool_calls[idx]["arguments"] += tc_delta.function.arguments

            # Check for finish_reason
            if choices[0].finish_reason:
                # Emit completed tool calls
                for idx, tc_data in sorted(active_tool_calls.items()):
                    yield {
                        "type": "tool_call",
                        "index": idx,
                        "name": tc_data["name"],
                        "arguments": tc_data["arguments"],
                        "call_id": tc_data["call_id"],
                    }
                active_tool_calls.clear()

        # Emit final usage
        cost = 0.0
        try:
            if accumulated_model and (accumulated_input_tokens or accumulated_output_tokens):
                prompt_cost, completion_cost = litellm.cost_per_token(
                    model=accumulated_model,
                    prompt_tokens=accumulated_input_tokens,
                    completion_tokens=accumulated_output_tokens,
                )
                cost = float(prompt_cost + completion_cost)
        except Exception:
            logger.debug("Could not extract streaming cost", exc_info=True)

        yield {
            "type": "usage",
            "input_tokens": accumulated_input_tokens,
            "output_tokens": accumulated_output_tokens,
            "cost": cost,
            "model": accumulated_model,
        }

    def resolve_model_params(
        self,
        agent_model: AgentModelConfig | None,
    ) -> tuple[str | None, float, int]:
        """Resolve model, temperature, max_tokens from agent config + gateway defaults."""
        model = None
        temperature = self._config.model.temperature
        max_tokens = self._config.model.max_tokens

        if agent_model:
            if agent_model.name:
                model = agent_model.name
            if agent_model.temperature is not None:
                temperature = agent_model.temperature
            if agent_model.max_tokens is not None:
                max_tokens = agent_model.max_tokens

        return model, temperature, max_tokens

    def _parse_response(self, response: ModelResponse) -> LLMResponse:
        """Parse a LiteLLM ModelResponse into our LLMResponse."""
        choices = response.choices
        text: str | None = None
        if choices and choices[0].message and choices[0].message.content:
            text = choices[0].message.content

        tool_calls = _parse_tool_calls(response)

        usage = response.usage
        input_tokens = usage.prompt_tokens if usage else 0
        output_tokens = usage.completion_tokens if usage else 0

        model = response.model or ""

        return LLMResponse(
            text=text,
            tool_calls=tool_calls,
            model=model,
            input_tokens=input_tokens or 0,
            output_tokens=output_tokens or 0,
            cost=_extract_cost(response),
        )

    async def close(self) -> None:
        """Clean up resources."""
        # Router doesn't have a close method, but we include this for future use
        pass



"""Tests for ExecutionJob model."""

from __future__ import annotations

import pytest

from agent_gateway.queue.models import ExecutionJob


def test_job_round_trip_json() -> None:
    """ExecutionJob serialises to JSON and back without data loss."""
    job = ExecutionJob(
        execution_id="exec-1",
        agent_id="agent-a",
        message="Hello",
        input={"key": "value"},
        timeout_ms=30000,
        output_schema={"type": "object"},
        enqueued_at="2026-02-18T10:00:00+00:00",
        retry_count=2,
    )

    json_str = job.to_json()
    restored = ExecutionJob.from_json(json_str)

    assert restored == job


def test_job_round_trip_minimal() -> None:
    """Minimal job (only required fields) round-trips."""
    job = ExecutionJob(
        execution_id="exec-2",
        agent_id="agent-b",
        message="Hi",
    )

    restored = ExecutionJob.from_json(job.to_json())

    assert restored.execution_id == "exec-2"
    assert restored.agent_id == "agent-b"
    assert restored.message == "Hi"
    assert restored.input is None
    assert restored.timeout_ms is None
    assert restored.retry_count == 0


def test_job_is_frozen() -> None:
    """ExecutionJob is immutable."""
    job = ExecutionJob(execution_id="x", agent_id="a", message="m")
    with pytest.raises(AttributeError):
        job.execution_id = "y"  # type: ignore[misc]



"""Rate limiting integration via slowapi (optional dependency)."""

from __future__ import annotations

import logging
from typing import Any

from starlette.requests import Request
from starlette.responses import JSONResponse

from agent_gateway.config import RateLimitConfig

logger = logging.getLogger(__name__)

try:
    from slowapi import Limiter
    from slowapi.errors import RateLimitExceeded
    from slowapi.middleware import SlowAPIASGIMiddleware
    from slowapi.util import get_remote_address

    HAS_SLOWAPI = True
except ImportError:  # pragma: no cover
    HAS_SLOWAPI = False


def _get_forwarded_key_func(request: Request) -> str:
    """Extract client IP from X-Forwarded-For header, falling back to remote address."""
    forwarded = request.headers.get("x-forwarded-for")
    if forwarded:
        # First IP in the chain is the original client
        return forwarded.split(",")[0].strip()
    return request.client.host if request.client else "127.0.0.1"


async def _rate_limit_exceeded_handler(request: Request, exc: Any) -> JSONResponse:
    """Return a 429 JSON response when the rate limit is exceeded."""
    return JSONResponse(
        status_code=429,
        content={"detail": "Rate limit exceeded"},
    )


class _StreamingSafeSlowAPIMiddleware:
    """Wraps SlowAPIASGIMiddleware to fix a bug where it re-sends http.response.start
    for every body chunk in streaming responses, violating the ASGI protocol."""

    def __init__(self, app: Any) -> None:
        self._inner = SlowAPIASGIMiddleware(app)

    async def __call__(self, scope: Any, receive: Any, send: Any) -> None:
        if scope["type"] != "http":
            await self._inner(scope, receive, send)
            return

        started = False

        async def deduped_send(message: Any) -> None:
            nonlocal started
            if message["type"] == "http.response.start":
                if started:
                    return  # Drop duplicate — slowapi streaming bug workaround
                started = True
            await send(message)

        await self._inner(scope, receive, deduped_send)


def setup_rate_limiting(
    app: Any,
    config: RateLimitConfig,
) -> tuple[Any, type[Any]] | None:
    """Configure rate limiting on the given application.

    Returns a ``(limiter, middleware_class)`` tuple if slowapi is available,
    or ``None`` if the optional dependency is missing.
    """
    if not HAS_SLOWAPI:
        logger.warning(
            "Rate limiting is enabled but 'slowapi' is not installed. "
            "Install with: pip install agents-gateway[rate-limiting]"
        )
        return None

    key_func = _get_forwarded_key_func if config.trust_forwarded_for else get_remote_address

    limiter = Limiter(
        key_func=key_func,
        default_limits=[config.default_limit],
        headers_enabled=True,
        storage_uri=config.storage_uri or None,
    )

    # slowapi requires the limiter on app.state
    app.state.limiter = limiter

    app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)

    return limiter, _StreamingSafeSlowAPIMiddleware



"""File-based memory backend using MEMORY.md per agent + per-user files.

Stores memories as structured markdown sections grouped by type.
Zero infrastructure — human-readable, git-committable, inspectable.

Layout:
    workspace/agents/{agent_id}/MEMORY.md              — agent-level (global) memories
    workspace/agents/{agent_id}/memory/{user_id}.md    — per-user memories
"""

from __future__ import annotations

import asyncio
import hashlib
import logging
import re
import uuid
from datetime import UTC, datetime
from pathlib import Path

from agent_gateway.memory.domain import (
    MemoryRecord,
    MemorySearchResult,
    MemorySource,
    MemoryType,
)

logger = logging.getLogger(__name__)

_SECTION_HEADING = re.compile(r"^## (Semantic|Episodic|Procedural)\s*$")
_MEMORY_LINE = re.compile(
    r"^- (?:\[([^\]]+)\] )?(.+)$"  # optional [id] prefix, then content
)

_SAFE_AGENT_ID = re.compile(r"^[a-zA-Z0-9][a-zA-Z0-9_-]*$")
_SAFE_FILENAME = re.compile(r"^[a-zA-Z0-9][a-zA-Z0-9_.@-]*$")

_TYPE_TO_HEADING: dict[MemoryType, str] = {
    MemoryType.SEMANTIC: "Semantic",
    MemoryType.EPISODIC: "Episodic",
    MemoryType.PROCEDURAL: "Procedural",
}
_HEADING_TO_TYPE: dict[str, MemoryType] = {v: k for k, v in _TYPE_TO_HEADING.items()}


def _user_filename(user_id: str) -> str:
    """Derive a safe filename from a user_id.

    If the user_id is already filesystem-safe (e.g. 'alice'), use it directly.
    Otherwise hash it to avoid path traversal or special characters.
    """
    if _SAFE_FILENAME.match(user_id) and len(user_id) <= 100:
        return f"{user_id}.md"
    hashed = hashlib.sha256(user_id.encode()).hexdigest()[:16]
    return f"user-{hashed}.md"


class FileMemoryRepository:
    """Repository that reads/writes structured markdown files.

    Supports both agent-level (global) and per-user memory files.

    .. note::
        All file I/O is synchronous (stdlib ``pathlib``/``open``).  This is
        intentional for the default zero-dependency backend: memory operations
        are infrequent and the files are small (< 100 KB).  For
        high-throughput scenarios, swap in an async backend (e.g. pgvector)
        via ``gateway.set_memory_backend()``.
    """

    def __init__(self, workspace_root: Path, max_lines: int = 200) -> None:
        self._root = workspace_root
        self._max_lines = max_lines
        self._locks: dict[str, asyncio.Lock] = {}

    def _lock_for(self, key: str) -> asyncio.Lock:
        if key not in self._locks:
            self._locks[key] = asyncio.Lock()
        return self._locks[key]

    def _agent_dir(self, agent_id: str) -> Path:
        if not _SAFE_AGENT_ID.match(agent_id):
            raise ValueError(f"Invalid agent_id: {agent_id!r}")
        path = self._root / "agents" / agent_id
        if not path.resolve().is_relative_to(self._root.resolve()):
            raise ValueError(f"Path traversal detected for agent_id: {agent_id!r}")
        return path

    def _global_memory_path(self, agent_id: str) -> Path:
        return self._agent_dir(agent_id) / "MEMORY.md"

    def _user_memory_path(self, agent_id: str, user_id: str) -> Path:
        agent_dir = self._agent_dir(agent_id)
        path = agent_dir / "memory" / _user_filename(user_id)
        if not path.resolve().is_relative_to(agent_dir.resolve()):
            raise ValueError(f"Path traversal detected for user_id: {user_id!r}")
        return path

    def _memory_path(self, agent_id: str, user_id: str | None = None) -> Path:
        if user_id is not None:
            return self._user_memory_path(agent_id, user_id)
        return self._global_memory_path(agent_id)

    def _lock_key(self, agent_id: str, user_id: str | None = None) -> str:
        if user_id is not None:
            return f"{agent_id}:user:{user_id}"
        return agent_id

    def _parse_file(
        self, path: Path, agent_id: str, user_id: str | None = None
    ) -> list[MemoryRecord]:
        """Parse a MEMORY.md file into MemoryRecord objects."""
        if not path.exists():
            return []

        content = path.read_text(encoding="utf-8")
        records: list[MemoryRecord] = []
        current_type = MemoryType.SEMANTIC
        now = datetime.now(UTC)

        for line in content.splitlines():
            heading_match = _SECTION_HEADING.match(line)
            if heading_match:
                current_type = _HEADING_TO_TYPE[heading_match.group(1)]
                continue

            mem_match = _MEMORY_LINE.match(line)
            if mem_match:
                memory_id = mem_match.group(1) or uuid.uuid4().hex[:8]
                memory_content = mem_match.group(2)
                records.append(
                    MemoryRecord(
                        id=memory_id,
                        agent_id=agent_id,
                        content=memory_content,
                        memory_type=current_type,
                        source=MemorySource.MANUAL,
                        user_id=user_id,
                        created_at=now,
                        updated_at=now,
                    )
                )

        return records

    def _write_records(self, path: Path, records: list[MemoryRecord]) -> None:
        """Write records to a memory file as structured markdown."""
        path.parent.mkdir(parents=True, exist_ok=True)

        grouped: dict[MemoryType, list[MemoryRecord]] = {
            MemoryType.SEMANTIC: [],
            MemoryType.EPISODIC: [],
            MemoryType.PROCEDURAL: [],
        }
        for r in records:
            grouped[r.memory_type].append(r)

        lines: list[str] = []
        for mem_type, heading in _TYPE_TO_HEADING.items():
            type_records = grouped[mem_type]
            if not type_records:
                continue
            lines.append(f"## {heading}")
            for r in type_records:
                lines.append(f"- [{r.id}] {r.content}")
            lines.append("")

        path.write_text("\n".join(lines), encoding="utf-8")

    async def save(self, record: MemoryRecord) -> None:
        key = self._lock_key(record.agent_id, record.user_id)
        path = self._memory_path(record.agent_id, record.user_id)
        async with self._lock_for(key):
            records = self._parse_file(path, record.agent_id, record.user_id)

            # Upsert: replace existing record with same id
            found = False
            for i, r in enumerate(records):
                if r.id == record.id:
                    records[i] = record
                    found = True
                    break

            if not found:
                records.append(record)

            self._write_records(path, records)

    async def get(self, agent_id: str, memory_id: str) -> MemoryRecord | None:
        # Search global first, then all user files
        global_records = self._parse_file(self._global_memory_path(agent_id), agent_id)
        for r in global_records:
            if r.id == memory_id:
                return r

        user_dir = self._agent_dir(agent_id) / "memory"
        if user_dir.exists():
            for user_file in user_dir.glob("*.md"):
                user_records = self._parse_file(user_file, agent_id)
                for r in user_records:
                    if r.id == memory_id:
                        return r
        return None

    async def list_memories(
        self,
        agent_id: str,
        *,
        user_id: str | None = None,
        include_global: bool = True,
        memory_type: MemoryType | None = None,
        limit: int = 50,
        offset: int = 0,
    ) -> list[MemoryRecord]:
        records: list[MemoryRecord] = []

        # Include user-specific memories
        if user_id is not None:
            user_path = self._user_memory_path(agent_id, user_id)
            records.extend(self._parse_file(user_path, agent_id, user_id))

        # Include global memories
        if include_global or user_id is None:
            global_path = self._global_memory_path(agent_id)
            records.extend(self._parse_file(global_path, agent_id))

        if memory_type is not None:
            records = [r for r in records if r.memory_type == memory_type]
        # Most recent first (by position in file, newest appended last)
        records.reverse()
        return records[offset : offset + limit]

    async def search(
        self,
        agent_id: str,
        query: str,
        *,
        user_id: str | None = None,
        include_global: bool = True,
        memory_type: MemoryType | None = None,
        limit: int = 10,
    ) -> list[MemorySearchResult]:
        """Simple keyword search — case-insensitive matching."""
        records: list[MemoryRecord] = []

        if user_id is not None:
            user_path = self._user_memory_path(agent_id, user_id)
            records.extend(self._parse_file(user_path, agent_id, user_id))

        if include_global or user_id is None:
            global_path = self._global_memory_path(agent_id)
            records.extend(self._parse_file(global_path, agent_id))

        if memory_type is not None:
            records = [r for r in records if r.memory_type == memory_type]

        query_lower = query.lower()
        query_words = query_lower.split()

        results: list[MemorySearchResult] = []
        for r in records:
            content_lower = r.content.lower()
            # Score: fraction of query words found in content
            matches = sum(1 for w in query_words if w in content_lower)
            if matches > 0:
                score = matches / len(query_words) if query_words else 0.0
                # Boost user-specific memories when searching with user context
                if user_id is not None and r.user_id == user_id:
                    score *= 1.2
                results.append(MemorySearchResult(record=r, score=score))

        results.sort(key=lambda x: x.score, reverse=True)
        return results[:limit]

    async def delete(self, agent_id: str, memory_id: str) -> bool:
        # Try global file first
        global_path = self._global_memory_path(agent_id)
        key = self._lock_key(agent_id)
        async with self._lock_for(key):
            records = self._parse_file(global_path, agent_id)
            original_count = len(records)
            records = [r for r in records if r.id != memory_id]
            if len(records) < original_count:
                self._write_records(global_path, records)
                return True

        # Try user files
        user_dir = self._agent_dir(agent_id) / "memory"
        if user_dir.exists():
            for user_file in user_dir.glob("*.md"):
                user_key = f"{agent_id}:file:{user_file.name}"
                async with self._lock_for(user_key):
                    user_records = self._parse_file(user_file, agent_id)
                    original_count = len(user_records)
                    user_records = [r for r in user_records if r.id != memory_id]
                    if len(user_records) < original_count:
                        self._write_records(user_file, user_records)
                        return True
        return False

    async def delete_all(self, agent_id: str, user_id: str | None = None) -> int:
        if user_id is not None:
            # Delete only this user's memories
            user_path = self._user_memory_path(agent_id, user_id)
            key = self._lock_key(agent_id, user_id)
            async with self._lock_for(key):
                records = self._parse_file(user_path, agent_id, user_id)
                count = len(records)
                if count > 0:
                    user_path.write_text("", encoding="utf-8")
                return count

        # Delete all memories (global + all user files)
        total = 0
        global_path = self._global_memory_path(agent_id)
        key = self._lock_key(agent_id)
        async with self._lock_for(key):
            records = self._parse_file(global_path, agent_id)
            total += len(records)
            if records:
                global_path.write_text("", encoding="utf-8")

        user_dir = self._agent_dir(agent_id) / "memory"
        if user_dir.exists():
            for user_file in user_dir.glob("*.md"):
                user_key = f"{agent_id}:file:{user_file.name}"
                async with self._lock_for(user_key):
                    user_records = self._parse_file(user_file, agent_id)
                    total += len(user_records)
                    if user_records:
                        user_file.write_text("", encoding="utf-8")

        return total

    async def count(self, agent_id: str, user_id: str | None = None) -> int:
        if user_id is not None:
            user_path = self._user_memory_path(agent_id, user_id)
            return len(self._parse_file(user_path, agent_id, user_id))

        # Count all (global + all user files)
        total = len(self._parse_file(self._global_memory_path(agent_id), agent_id))
        user_dir = self._agent_dir(agent_id) / "memory"
        if user_dir.exists():
            for user_file in user_dir.glob("*.md"):
                total += len(self._parse_file(user_file, agent_id))
        return total


class FileMemoryBackend:
    """File-based memory backend using MEMORY.md per agent + per-user files.

    Zero-config default. Stores memories as structured markdown sections
    grouped by type. Supports keyword search and per-agent asyncio locks.
    """

    def __init__(self, workspace_root: str | Path, max_lines: int = 200) -> None:
        self._root = Path(workspace_root)
        self._max_lines = max_lines
        self._repo = FileMemoryRepository(self._root, max_lines)

    async def initialize(self) -> None:
        pass

    async def dispose(self) -> None:
        pass

    @property
    def memory_repo(self) -> FileMemoryRepository:
        return self._repo



"""Add missing indexes on created_at and status columns.

Revision ID: 007
Revises: 006
Create Date: 2026-02-25
"""

from collections.abc import Sequence

from alembic import op

revision: str = "007"
down_revision: str | None = "006"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    op.create_index("ix_executions_created_at", "executions", ["created_at"])
    op.create_index("ix_executions_status", "executions", ["status"])
    op.create_index("ix_conversations_created_at", "conversations", ["created_at"])
    op.create_index("ix_audit_log_created_at", "audit_log", ["created_at"])


def downgrade() -> None:
    op.drop_index("ix_audit_log_created_at", table_name="audit_log")
    op.drop_index("ix_conversations_created_at", table_name="conversations")
    op.drop_index("ix_executions_status", table_name="executions")
    op.drop_index("ix_executions_created_at", table_name="executions")



"""Tests for distributed lock backends."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock

import pytest

from agent_gateway.exceptions import SchedulerLockError
from agent_gateway.scheduler.lock import (
    NullDistributedLock,
    PostgresDistributedLock,
    RedisDistributedLock,
)


class TestNullDistributedLock:
    async def test_acquire_always_returns_true(self) -> None:
        lock = NullDistributedLock()
        assert await lock.acquire("test-lock", 300) is True

    async def test_release_is_noop(self) -> None:
        lock = NullDistributedLock()
        await lock.release("test-lock")  # should not raise

    async def test_initialize_and_dispose(self) -> None:
        lock = NullDistributedLock()
        await lock.initialize()
        await lock.dispose()


class TestRedisDistributedLock:
    async def test_acquire_succeeds_first_time(self) -> None:
        lock = RedisDistributedLock(url="redis://localhost:6379/0")
        mock_redis = AsyncMock()
        mock_redis.set.return_value = True
        lock._redis = mock_redis

        result = await lock.acquire("test-lock", 300)
        assert result is True
        mock_redis.set.assert_called_once_with("ag:sched-lock:test-lock", "1", nx=True, ex=300)

    async def test_acquire_fails_when_held(self) -> None:
        lock = RedisDistributedLock(url="redis://localhost:6379/0")
        mock_redis = AsyncMock()
        mock_redis.set.return_value = None
        lock._redis = mock_redis

        result = await lock.acquire("test-lock", 300)
        assert result is False

    async def test_release_deletes_key(self) -> None:
        lock = RedisDistributedLock(url="redis://localhost:6379/0")
        mock_redis = AsyncMock()
        lock._redis = mock_redis

        await lock.release("test-lock")
        mock_redis.delete.assert_called_once_with("ag:sched-lock:test-lock")

    async def test_acquire_raises_when_not_initialized(self) -> None:
        lock = RedisDistributedLock(url="redis://localhost:6379/0")
        with pytest.raises(SchedulerLockError, match="not initialized"):
            await lock.acquire("test-lock", 300)

    async def test_release_noop_when_not_initialized(self) -> None:
        lock = RedisDistributedLock(url="redis://localhost:6379/0")
        await lock.release("test-lock")  # should not raise

    async def test_dispose_closes_redis(self) -> None:
        lock = RedisDistributedLock(url="redis://localhost:6379/0")
        mock_redis = AsyncMock()
        lock._redis = mock_redis

        await lock.dispose()
        mock_redis.aclose.assert_called_once()
        assert lock._redis is None

    async def test_custom_key_prefix(self) -> None:
        lock = RedisDistributedLock(url="redis://localhost:6379/0", key_prefix="custom:")
        mock_redis = AsyncMock()
        mock_redis.set.return_value = True
        lock._redis = mock_redis

        await lock.acquire("my-lock", 60)
        mock_redis.set.assert_called_once_with("custom:my-lock", "1", nx=True, ex=60)


class TestPostgresDistributedLock:
    async def test_acquire_succeeds(self) -> None:
        mock_engine = AsyncMock()
        mock_conn = AsyncMock()
        mock_result = MagicMock()
        mock_result.scalar.return_value = True
        mock_conn.execute.return_value = mock_result
        mock_engine.connect.return_value = mock_conn

        lock = PostgresDistributedLock(engine=mock_engine)
        result = await lock.acquire("test-lock", 300)

        assert result is True
        assert "test-lock" in lock._held_connections
        mock_conn.close.assert_not_called()

    async def test_acquire_fails_when_held(self) -> None:
        mock_engine = AsyncMock()
        mock_conn = AsyncMock()
        mock_result = MagicMock()
        mock_result.scalar.return_value = False
        mock_conn.execute.return_value = mock_result
        mock_engine.connect.return_value = mock_conn

        lock = PostgresDistributedLock(engine=mock_engine)
        result = await lock.acquire("test-lock", 300)

        assert result is False
        assert "test-lock" not in lock._held_connections
        mock_conn.close.assert_called_once()

    async def test_release_unlocks_and_closes_connection(self) -> None:
        mock_conn = AsyncMock()
        lock = PostgresDistributedLock(engine=AsyncMock())
        lock._held_connections["test-lock"] = mock_conn

        await lock.release("test-lock")
        # Should call pg_advisory_unlock then close
        assert mock_conn.execute.call_count == 1
        call_args = mock_conn.execute.call_args
        text_clause = call_args[0][0]
        assert "pg_advisory_unlock" in text_clause.text
        mock_conn.close.assert_called_once()
        assert "test-lock" not in lock._held_connections

    async def test_release_noop_when_not_held(self) -> None:
        lock = PostgresDistributedLock(engine=AsyncMock())
        await lock.release("nonexistent")  # should not raise

    async def test_dispose_closes_all_held(self) -> None:
        conn1 = AsyncMock()
        conn2 = AsyncMock()
        lock = PostgresDistributedLock(engine=AsyncMock())
        lock._held_connections = {"lock-a": conn1, "lock-b": conn2}

        await lock.dispose()
        conn1.close.assert_called_once()
        conn2.close.assert_called_once()
        assert len(lock._held_connections) == 0
